cdef extern from "<string_view>" namespace "std" nogil:
    cppclass string_view:
        string_view() except +
        string_view(const char *, size_t) except +
        const char *data()
        size_t size()



from athenian.api.native.mi_heap_destroy_stl_allocator cimport mi_heap_destroy_stl_allocator


cdef extern from "chunked_stream.h" nogil:
    cdef cppclass chunked_stream[I=*]:
        chunked_stream chunked_stream[X](mi_heap_destroy_stl_allocator[X] &) except +
        void write(const void *buffer, size_t size)
        size_t dump(char *output, size_t output_size)
        size_t size()



from importlib.machinery import SourceFileLoader
import os
from pathlib import Path
import site

site.ENABLE_USER_SITE = True  # workaround https://github.com/pypa/pip/issues/7953

# The following import has to stay after imports from `setuptools`:
# - https://stackoverflow.com/questions/21594925/
#     error-each-element-of-ext-modules-option-must-be-an-extension-instance-or-2-t
from Cython.Build import cythonize  # noqa: E402
import numpy as np  # noqa: E402
from setuptools import find_packages, setup  # noqa: E402

project_root = Path(__file__).parent
code_root = project_root / "medvedi"
os.chdir(str(project_root))
version = SourceFileLoader("version", str(code_root / "metadata.py")).load_module()

with open(project_root / "README.md", encoding="utf-8") as f:
    long_description = f.read()


ext_modules = cythonize(
    [
        str(path)
        # fmt: off
        for path in (
            code_root / "accelerators.pyx",
            code_root / "io.pyx",
            code_root / "native" / "mi_heap_destroy_stl_allocator.pyx",
        )
        # fmt: on
    ],
)
for ext_mod in ext_modules:
    ext_mod.library_dirs = [str(code_root)]


setup(
    name=version.__package__.replace(".", "-"),
    description=version.__description__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=version.__version__,
    license="Apache 2.0",
    author="Athenian",
    author_email="vadim@athenian.co",
    url="https://github.com/athenianco/medvedi",
    download_url="https://github.com/athenianco/medvedi",
    packages=find_packages(exclude=["*tests"]),
    ext_modules=ext_modules,
    include_dirs=[np.get_include(), str(code_root)],
    keywords=[],
    install_requires=["numpy>=1.23,<1.24"],
    extras_require={
        "arrow": ["pyarrow"],
    },
    tests_require=[],  # see requirements-test.txt
    package_data={
        "": ["*.md"],
        "medvedi": ["../requirements.txt", "libmimalloc.so*", "mimalloc.h", "*.pyx"],
        "medvedi.native": ["*.pyx", "*.pxd", "*.h"],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)



import numpy as np

mergeable_dtype_kinds = frozenset(("i", "u", "m", "M", "S"))


def merge_to_str(*arrs: np.ndarray) -> np.ndarray:
    """
    Convert one or more arrays of integers, bytes, datetime64, timedelta64 to \
    S(total_bytes + (last array is not "S")).

    We cannot use arr.byteswap().view("S8") because the trailing zeros get discarded \
    in np.char.add. Thus, we have to pad with ";" unless the last array is bytes.

    We copy bytes ("S"), datetime64, timedelta64 verbatim.
    """
    assert len(arrs) > 0
    size = len(arrs[0])
    lock = arrs[-1].dtype.kind != "S"
    str_len = sum(arr.dtype.itemsize for arr in arrs) + lock
    arena = np.full((size, str_len), ord(";") if lock else 0, dtype=np.uint8)
    pos = 0
    for arr in arrs:
        kind = arr.dtype.kind
        if kind not in mergeable_dtype_kinds:
            raise ValueError(
                f"array's dtype.kind {arr.dtype} must be one of "
                f"{', '.join(mergeable_dtype_kinds)}",
            )
        if len(arr) != size:
            raise ValueError(f"all arrays must have the same length: {len(arr)} != {size}")
        itemsize = arr.dtype.itemsize
        if itemsize > 1 and kind != "S":
            col = arr.byteswap()
        else:
            col = arr
        arena[:, pos : pos + itemsize] = col.view(np.uint8).reshape(size, itemsize)
        pos += itemsize
    return arena.ravel().view(f"S{str_len}")



from cpython cimport PyObject

ctypedef PyObject *PyObjectPtr

cdef extern from "structmember.h":
    ctypedef struct PyMemberDef:
        const char *name
        int type
        Py_ssize_t offset
        int flags
        const char *doc

cdef extern from "Python.h":
    ctypedef PyObject *(*allocfunc)(PyTypeObject *cls, Py_ssize_t nitems)

    ctypedef struct PyTypeObject:
        allocfunc tp_alloc
        PyMemberDef *tp_members

    bint PyObject_TypeCheck(PyObject *, PyTypeObject *) nogil
    PyTypeObject *Py_TYPE(const PyObject *) nogil

    bint PyLong_CheckExact(PyObject *) nogil
    long PyLong_AsLong(PyObject *) nogil

    double PyFloat_AS_DOUBLE(PyObject *) nogil
    bint PyFloat_CheckExact(PyObject *) nogil
    bint PyFloat_Check(PyObject *) nogil

    PyObject *PyList_New(Py_ssize_t len)
    bint PyList_CheckExact(PyObject *) nogil
    Py_ssize_t PyList_GET_SIZE(PyObject *) nogil
    PyObject *PyList_GET_ITEM(PyObject *, Py_ssize_t) nogil
    void PyList_SET_ITEM(PyObject *list, Py_ssize_t i, PyObject *o) nogil

    PyObject *PyTuple_GET_ITEM(PyObject *, Py_ssize_t) nogil

    bint PyDict_CheckExact(PyObject *) nogil
    int PyDict_Next(PyObject *p, Py_ssize_t *ppos, PyObject **pkey, PyObject **pvalue) nogil
    Py_ssize_t PyDict_Size(PyObject *p) nogil

    bint PyUnicode_Check(PyObject *) nogil
    Py_ssize_t PyUnicode_GET_LENGTH(PyObject *) nogil
    unsigned int PyUnicode_KIND(PyObject *) nogil
    void *PyUnicode_DATA(PyObject *) nogil
    Py_ssize_t PyUnicode_FindChar(
        PyObject *str, Py_UCS4 ch, Py_ssize_t start, Py_ssize_t end, int direction
    ) nogil

    bint PyBytes_Check(PyObject *) nogil
    char *PyBytes_AS_STRING(PyObject *) nogil
    Py_ssize_t PyBytes_GET_SIZE(PyObject *) nogil

    bint PyByteArray_CheckExact(PyObject *) nogil
    char *PyByteArray_AS_STRING(PyObject *) nogil

    unsigned int PyUnicode_1BYTE_KIND
    unsigned int PyUnicode_2BYTE_KIND
    unsigned int PyUnicode_4BYTE_KIND

    PyObject *Py_None
    PyObject *Py_True
    PyObject *Py_False

    PyTypeObject PyLong_Type
    PyTypeObject PyFloat_Type
    PyTypeObject PyUnicode_Type
    PyTypeObject PyBool_Type
    PyTypeObject PyList_Type
    PyTypeObject PyDict_Type
    PyTypeObject PyBaseObject_Type

    void Py_INCREF(PyObject *)
    void Py_DECREF(PyObject *)

    object PyUnicode_FromStringAndSize(const char *, Py_ssize_t)
    str PyUnicode_FromKindAndData(unsigned int kind, void *buffer, Py_ssize_t size)
    str PyUnicode_FromString(const char *)
    object PyUnicode_New(Py_ssize_t, Py_UCS4)
    PyObject *PyBytes_FromStringAndSize(char *v, Py_ssize_t len)
    PyObject *PyLong_FromLong(long v)
    PyObject *PyObject_GetItem(PyObject *o, PyObject *key)
    PyObject *PyObject_GetAttr(PyObject *o, PyObject *attr_name)

cdef extern from "datetime.h" nogil:
    bint PyDateTime_Check(PyObject *)
    bint PyDelta_Check(PyObject *)

    int PyDateTime_GET_YEAR(PyObject *)
    int PyDateTime_GET_MONTH(PyObject *)
    int PyDateTime_GET_DAY(PyObject *)

    int PyDateTime_DATE_GET_HOUR(PyObject *)
    int PyDateTime_DATE_GET_MINUTE(PyObject *)
    int PyDateTime_DATE_GET_SECOND(PyObject *)

    int PyDateTime_DELTA_GET_DAYS(PyObject *)
    int PyDateTime_DELTA_GET_SECONDS(PyObject *)

    PyObject *PyDateTime_DATE_GET_TZINFO(PyObject *)

    ctypedef struct PyDateTime_CAPI:
        PyObject *TimeZone_UTC
